package net.sourceforge.jetris;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;

import org.sat4j.BasicLauncher;
import org.sat4j.minisat.SolverFactory;
import org.sat4j.specs.ISolver;

public class Solver {
    public static final boolean DEBUG = true;
    public static final String  LEVEL_INPUT = "levels.txt";
    // sat4j or minisat? minisat is faster, but needs to be compiled
    public static final boolean SAT4JAVA = true; 

    /**
     * The cnf clauses storage before it is written to the file.
     * Needed since the number of clauses appears in the dimacs header.
     */
    private StringBuilder cnfWriter = new StringBuilder();

    /**
     * The counter of clauses, which is used for the dimacs header.
     */
    private int cnfCount = 0;

    
    /**
     * The counter of variables, which is used for the dimacs header.
     *
     * The variable sequence is W H#1 C#1 H#2 C#2 ... H#depth C#depth  bottom
     * Where H#i is H_1,1 H_2,1 ... H_width,1 ... H_1,height ... H_width,height
     */
    private int vars;

    /**
     * The index of a variable that is always false. This is guaranteed 
     * by a <code>"-<i>bottom</i> 0"</code> clause. This variable is
     * substituted for all variables that point to a field outside of the
     * level.
     */
    private int bottom;
	private int width;
	private int height;
	private int depth;
	private Level lvl;
    
	private final int numBlocks = 7;
	private final int numRot = 4;
	
    // ######## Complete the methods init() and move() ###############
    // ######## You _can_ leave everything else as is  ###############
    

    public Solver(Level lvl) {
    	this.lvl = lvl;
        this.width = lvl.getWidth();
        this.height = lvl.getHeight();
        this.depth = lvl.getSteps();
        
        //      ------- placement of blocks and cells
        //      |                    /---------- number of different blocks
        //		v				    /          /-------- different rotations of a block
        vars = (2*(width * height)+numBlocks+numRot)*(depth+1)+1; // the last one is "bottom"
        bottom = vars;
        comment("Bottom literal");
        clause(-bottom);
        comment("Problem clauses");

        init();
        move();
        writeout();
        solve();
        interpret();
    }


    private void solve() {
	try {
      if (SAT4JAVA) {
		BasicLauncher<ISolver> launcher = new BasicLauncher<ISolver>(
                    SolverFactory.instance());
		PrintWriter pw = new PrintWriter("out.ms");
		launcher.setLogWriter(pw);
        	launcher.run(new String[] { "in.ms" });
		launcher.displayResult();
		pw.flush();
		pw.close();
	    } else {
		    Process p = Runtime.getRuntime().exec(
		        "minisat in.ms out.ms");
		    p.waitFor();
	    }
	} catch(IOException ioe) {
	    throw new RuntimeException(ioe);
	} catch(InterruptedException ie) {
	    throw new RuntimeException(ie);
	}
    }
    
    static LinkedList<Action> solutionKeys;
   
	private void interpret() {
		solutionKeys = new LinkedList<Action>();
		
		try {
			BufferedReader msr = new BufferedReader(new FileReader("out.ms"));
			String res = msr.readLine();
			while (res.startsWith("c"))
				res = msr.readLine(); // skip comments
			if ("s SATISFIABLE".equals(res))
				res = "SAT"; // translate sat4j
			if (DEBUG)
				System.out.println("SOLVER SAYS:\n" + res);
			if (res.equals("SAT")) {
				// Split into single variables; remove sat4j's 'v' flag
				String[] model = msr.readLine().replace("v ", "").split(" ");
				int i = 0, t = 0, x = 0, y = 0;
				
				boolean blockPlaced=false;
				int varsRead=0;

				if (DEBUG) System.out.println("t = " + t + " -------------");
				
				for (String l : model) {					
					int val = Integer.parseInt(l);
					i++;
					varsRead++;
					
					if (i >= Figure.I && i <= Figure.Z && val > 0)
						if (DEBUG) System.out.println("Current Block: " + Figure.figureRep[i-1]);
					
					if (i == numBlocks + 1)
						if (DEBUG) System.out.println("Block Placement: ");

					//now the "tetromino-placement" block
					if (i > numBlocks && i <= numBlocks + width*height) {							
						if (val > 0) {
							blockPlaced = true;
							x = (i-8)%width;
							y = (i-8)/width;
							if (DEBUG) System.out.print("+");
						}
						else if (DEBUG) System.out.print(".");
						//one row is completed
						if ((i - 7) % width == 0)
							if (DEBUG) System.out.println();
					}
						
					if (i > numBlocks + width*height && i <= numBlocks+width*height + numRot) {
						if (val > 0) {
							int rot = (i-(numBlocks + width*height));
							if (DEBUG) System.out.println("Rotation: " + rot);
							
							if (blockPlaced) {
							switch(rot) {
							case Figure.Left:
								solutionKeys.add(Action.ROTATE);
								break;
							case Figure.Right:
								solutionKeys.add(Action.ROTATE);
								solutionKeys.add(Action.ROTATE);
								solutionKeys.add(Action.ROTATE);
								break;
							case Figure.Down:
								break;
							case Figure.Up:
								solutionKeys.add(Action.ROTATE);
								solutionKeys.add(Action.ROTATE);
								break;
							}
							
							int mid = width/2;
							int currx = mid;
							int curry = 0;
							
							while (currx != x) {
								if (x < mid) {
									solutionKeys.add(Action.LEFT);
									currx--;
								} else if (x > mid) {
									solutionKeys.add(Action.RIGHT);
									currx++;
								}
							}
							
							while (curry != y) {
								solutionKeys.add(Action.DOWN);
							    curry++;
							}
							
							solutionKeys.add(Action.DOWN);
						}						
					}
						}
									
					
					if (i == numBlocks + width*height+numRot+1)
						if (DEBUG) System.out.println("Cells: ");
					
					//now the "cell" block
					if (i > numBlocks + width*height+numRot && i <= numBlocks+2*width*height + numRot) {
						if (val > 0) {
							if (DEBUG) System.out.print("+");
						}
						else if (DEBUG) System.out.print(".");
						//one row is completed
						if ((i - (numBlocks + width*height+numRot)) % width == 0)
							if (DEBUG) System.out.println();
					}

					//if one block has been read...
					if (i == 2 * width * height + numBlocks + numRot) {
						i = 0;
						t++;
						if (DEBUG) System.out.println("t = " + t + " -------------");
						blockPlaced=false;
					}
					
				}
			} else if (DEBUG)
				System.out.println("UNSAT");
		} catch (IOException ioe) {
			throw new RuntimeException(ioe);
		}
		
		if (DEBUG)
			System.out.println("\nSOLUTION:" + solutionKeys);
	}

	//type of current block at time "step"
	//step: 0..depth
    private int B(int type, int step) {
    	if (type < Figure.I || type > Figure.Z) return bottom;
    		//2 * width * height + 7 + 4 variables per step
            return (2 * width * height + numBlocks + numRot)*(step)+type;
    }

    //true iff block is placed at x,y at time "step"
    private int P(int x, int y, int step) {
    	if (x < 1 || x > width || y < 1 || y > height) return bottom;
            return (2 * width * height + numBlocks + numRot)*(step)+x+(y-1)*width+numBlocks;
    }
    
    //rotation at time "step"
    private int R(int direction, int step) {
    	if (direction < Figure.Left || direction > Figure.Down) return bottom;
            return (2 * width * height + numBlocks + numRot)*(step)
            	+ numBlocks + width * height + direction;
    }
    
	//C(x,y,step) is true iff the cell at pos x,y at time "step" is filled
    private int C(int x, int y, int step) {
	if (x < 1 || x > width || y < 1 || y > height) return bottom;
		return (2 * width * height + numBlocks + numRot)*(step)
				+ numBlocks + numRot + width * height +x+(y-1)*width;
    }
    


    /**
     * given a number of indices, concatenate them and append a 0 to produce a
     * dimacs clause which is added to {@link #cnfWriter} and {@link #cnfCount}
     * incremented.
     * 
     * @param vars
     *            one or more int expressions (clause literals)
     */
    private void clause(int... vars) {
	    for (int i = 0; i < vars.length; i++) {
		    if (vars[i] != 0)
			    cnfWriter.append(vars[i]).append(" ");
	    }
	    cnfWriter.append("0\n");
	    cnfCount++;
    }

    
    private void comment(String s) {
        if (SAT4JAVA) return; // sat4java does not like to read comments
        cnfWriter.append("c "+s+"\n");
    }


    private void writeout() {
        try {
            FileWriter fw = new FileWriter("in.ms");
            fw.write("p cnf " + vars + " " + cnfCount + "\n");
            fw.write(""+cnfWriter);
            fw.close();
            if (DEBUG) System.out.println(
                "Done writing dimacs file, calling solver.");
        } catch(IOException ioe) {
            throw new RuntimeException(ioe);
        }
    }
    
   
    // Complete me
    private void init() {	
        comment("*** end init");
    }


    // Complete me
	private void move() {
//		for (int t = 0; t <= depth; t++) {
//			for (int y = 1; y <= height; y++) {
//				for (int x = 1; x <= width; x++) { 
//					//TODO
//				}
//				
//			}
//			
//		}
			
	}
}
